//
// Sample for stencil shadow volumes (z-fail) in OpenGL
//
// Author: Alex V. Boreskoff <alexboreskoff@mtu-net.ru>, <steps3d@narod.ru>
//

#include    "libExt.h"

#include    <glut.h>
#include    <stdio.h>
#include    <stdlib.h>
#include    <assert.h>

#include    "libTexture.h"
#include    "Torus.h"

Vector3D    eye   ( 7, 5, 7 );           // camera position
Vector4D    light ( 2, 0, 3, 1 );        // light position
unsigned    ambMap;                     // ambient texture
unsigned    diffuseMap;                 // diffuse texture
float       angle = 0;
Torus       torus ( 0.6, 1.5, 30, 30 );

/////////////////////////// Some useful shadowing operations ////////////////////

void    projectVertices ( Vertex * vertices, int numVertices, const Vector4D& light )
{
    for ( int i = 0; i < numVertices; i++ )
    {
        Vector4D& v = vertices [i].pos;

        vertices [i].inf.x = v.x * light.w - v.w * light.x;
        vertices [i].inf.y = v.y * light.w - v.w * light.y;
        vertices [i].inf.z = v.z * light.w - v.w * light.z;
        vertices [i].inf.w = 0;
    }
}

void    classifyFaces ( Vertex * vertices, Face * faces, int numFaces, const Vector4D& light )
{
   for ( int i = 0; i < numFaces; i++ )
    {
                                            // get vertex from triangular face
        Vertex& a = vertices [faces [i].index [0]];

                                            // check its orientation to the light
        Vector4D    l    = (light/light.w) - (a.pos/a.pos.w);
        float       dot = Vector3D ( l.x, l.y, l.z ) & faces [i].n;

        faces [i].frontFacing = (dot > 0);
    }
}

void    infFrustum ( float left, float right, float bottom, float top, float zNear, float zFar )
{
    float   proj [16];                      // new projection matrix

    proj [0 ] = 2*zNear/(right - left);         // P (0,0)
    proj [1 ] = 0;                              // P (0,1)
    proj [2 ] = 0;                              // P (0,2)
    proj [3 ] = 0;                              // P (0,3)

    proj [4 ] = 0;                              // P (1,0)
    proj [5 ] = 2*zNear/(top - bottom);         // P (1,1)
    proj [6 ] = 0;                              // P (1,2)
    proj [7 ] = 0;                              // P (1,3)

    proj [8 ] = (right+left)/(right - left);    // P (2,0)
    proj [9 ] = (top+bottom)/(top-bottom);      // P (2,1)
    proj [10] = -1;                             // P (2,2)
    proj [11] = -1;                             // P (2,3)

    proj [12] = 0;                              // P (3,0)
    proj [13] = 0;                              // P (3,1)
    proj [14] = -2*zNear;                       // P (3,2)
    proj [15] = 0;                              // P (3,3)

    glMultMatrixf ( proj );
}

void    drawShadowVolume ( Face * faces, int numFaces, Edge * edges, int numEdges, Vertex * vertices, int numVertices )
{
    int i;
                                            // draw extruded silhouette edges
    glBegin ( GL_QUADS );

    for ( i = 0; i < numEdges; i++ )
    {
        assert ( edges[i].f1 != -1 );
        assert ( edges[i].f2 != -1 );

        if ( faces [edges[i].f1].frontFacing == faces [edges[i].f2].frontFacing )
            continue;
                                            // we've found silhoette edge
                                            // extrude it
                                            // Note, use correct order for AB from front-facing facet
        int ia, ib;

        if ( faces [edges[i].f1].frontFacing )
        {
            if ( edges [i].rev1 )			// whether order of points differs from f1
            {
                ia = edges [i].b;
                ib = edges [i].a;
            }
            else
            {
                ia = edges [i].a;
                ib = edges [i].b;
            }
        }
        else								// faces [edges [i].f2] is frontFacing
        {
            if ( edges [i].rev2 )
            {
                ia = edges [i].b;
                ib = edges [i].a;
            }
            else
            {
                ia = edges [i].a;
                ib = edges [i].b;
            }
        }

        Vertex& a = vertices [ia];
        Vertex& b = vertices [ib];

                                            // draw extruded AB
        glVertex4fv ( b.pos );
        glVertex4fv ( a.pos );
        glVertex4fv ( a.inf );
        glVertex4fv ( b.inf );
    }

    glEnd   ();

                                            // now draw front-facing the light faces & their projections
    glBegin ( GL_TRIANGLES );

    for ( i = 0; i < numFaces; i++ )
    {
        Vertex& a = vertices [faces [i].index [0]];
        Vertex& b = vertices [faces [i].index [1]];
        Vertex& c = vertices [faces [i].index [2]];


        if ( faces [i].frontFacing )
        {
                                            // draw ABC
            glVertex4fv ( a.pos );
            glVertex4fv ( b.pos );
            glVertex4fv ( c.pos );
        }
        else
        {                                    // draw ABC's projection
            glVertex4fv ( a.inf );
            glVertex4fv ( b.inf );
            glVertex4fv ( c.inf );
        }
    }

    glEnd ();
}

/////////////////////////////////////////////////////////////////////////////////

void init ()
{
    glClearColor   ( 0, 0, 0, 1 );
    glClearStencil ( 0 );
    glEnable       ( GL_DEPTH_TEST );
    glDepthFunc    ( GL_LESS );

    glHint ( GL_POLYGON_SMOOTH_HINT,         GL_NICEST );
    glHint ( GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST );
}

void display ()
{
    glClear ( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT );

                                                // draw unlit geometry
    glMatrixMode       ( GL_MODELVIEW );
    glColor3f          ( 0.5, 0.5, 0.5 );
    glActiveTextureARB ( GL_TEXTURE0_ARB );
    glEnable           ( GL_TEXTURE_2D );
    glBindTexture      ( GL_TEXTURE_2D, diffuseMap );

    torus.drawUnlit ();

                                                // draw unlit floor
    glActiveTextureARB ( GL_TEXTURE0_ARB );
    glEnable           ( GL_TEXTURE_2D );
    glBindTexture      ( GL_TEXTURE_2D, ambMap );

    glBegin ( GL_QUADS );

    glTexCoord2f ( 0, 0 );
    glVertex3f   ( -8, -8, -1 );

    glTexCoord2f ( 1, 0 );
    glVertex3f   ( 8, -8, -1 );

    glTexCoord2f ( 1, 1 );
    glVertex3f   ( 8, 8, -1 );

    glTexCoord2f ( 0, 1 );
    glVertex3f   ( -8, 8, -1 );

    glEnd   ();

                                                // draw shadow volumes
    projectVertices ( torus.getVertices (), torus.getNumVertices (), light );
    classifyFaces   ( torus.getVertices (), torus.getFaces (), torus.getNumFaces (), light );

                                                // draw back faces
    glColorMask   ( GL_FALSE, GL_FALSE, GL_FALSE, GL_FALSE );
    glDepthMask   ( GL_FALSE );
    glDisable     ( GL_TEXTURE_2D );
    glEnable      ( GL_STENCIL_TEST );
    glStencilMask ( ~0 );
    glStencilFunc ( GL_ALWAYS, 0, ~0 );

    glEnable      ( GL_CULL_FACE );
    glCullFace    ( GL_FRONT );

    glStencilOp   ( GL_KEEP,                    // stencil test failed
                    GL_INCR,                    // depth test failed
                    GL_KEEP );                  // depth test passed


    drawShadowVolume ( torus.getFaces (), torus.getNumFaces (), torus.getEdges (), torus.getNumEdges (), torus.getVertices (), torus.getNumVertices () );

                                                // draw front faces
    glCullFace    ( GL_BACK );

    glStencilOp   ( GL_KEEP,                    // stencil test failed
                    GL_DECR,                    // depth test failed
                    GL_KEEP );                  // depth test passed

    drawShadowVolume ( torus.getFaces (), torus.getNumFaces (), torus.getEdges (), torus.getNumEdges (), torus.getVertices (), torus.getNumVertices () );

    glColorMask ( GL_TRUE, GL_TRUE, GL_TRUE, GL_TRUE );
    glDepthMask ( GL_TRUE );
    glDisable   ( GL_STENCIL_TEST );
    glDisable   ( GL_CULL_FACE );
    glColor4f   ( 1, 1, 1, 1 );

                                                // draw the light
    glMatrixMode ( GL_MODELVIEW );
    glPushMatrix ();

    glTranslatef       ( light.x, light.y, light.z );
    glActiveTextureARB ( GL_TEXTURE0_ARB );
    glDisable          ( GL_TEXTURE_2D );
    glutSolidSphere    ( 0.1f, 15, 15 );
    glPopMatrix        ();

                                                // draw lighted objects
    glEnable      ( GL_TEXTURE_2D );
    glEnable      ( GL_STENCIL_TEST );
    glDepthFunc   ( GL_EQUAL );
    glStencilFunc ( GL_EQUAL, 0, ~0 );          // pass on stencil == 0
    glStencilOp   ( GL_KEEP, GL_KEEP, GL_KEEP );

                                                // draw floor
    glActiveTextureARB ( GL_TEXTURE0_ARB );
    glEnable           ( GL_TEXTURE_2D );
    glBindTexture      ( GL_TEXTURE_2D, ambMap );

    glBegin ( GL_QUADS );

    glTexCoord2f ( 0, 0 );
    glVertex3f   ( -8, -8, -1 );

    glTexCoord2f ( 1, 0 );
    glVertex3f   ( 8, -8, -1 );

    glTexCoord2f ( 1, 1 );
    glVertex3f   ( 8, 8, -1 );

    glTexCoord2f ( 0, 1 );
    glVertex3f   ( -8, 8, -1 );

    glEnd   ();
                                                // setup texture units
                                                // bind bump (normal) map to texture unit 0
    glActiveTextureARB ( GL_TEXTURE0_ARB );
    glEnable           ( GL_TEXTURE_2D );
    glBindTexture      ( GL_TEXTURE_2D, diffuseMap );

    torus.draw  ();

    glDisable   ( GL_STENCIL_TEST );
    glDepthFunc ( GL_LESS );

    glutSwapBuffers ();
}

void reshape ( int w, int h )
{
    float   fov    = 60;
    float   zNear  = 1;
    float   zFar   = 60;
    float   aspect = (float) w / (float) h;
    float   top    = zNear * tan ( 0.5 * fov * M_PI / 180 );
    float   bottom = -top;
    float   right  = top * aspect;
    float   left   = -right;

    glViewport     ( 0, 0, (GLsizei)w, (GLsizei)h );
    glMatrixMode   ( GL_PROJECTION );
    glLoadIdentity ();

    infFrustum ( left, right, bottom, top, zNear, zFar );

    glMatrixMode   ( GL_MODELVIEW );
    glLoadIdentity ();
    gluLookAt      ( eye.x, eye.y, eye.z,    // eye
                     0, 0, 0,                // center
                     0, 0, 1 );              // up
}

void key ( unsigned char key, int x, int y )
{
    if ( key == 27 || key == 'q' || key == 'Q' )        // quit requested
        exit ( 0 );
}

void    animate ()
{
    angle  = 0.004f * glutGet ( GLUT_ELAPSED_TIME );

    light.x = 2*cos ( angle );
    light.y = 2*sin ( angle );
    light.z = 3 + 0.9 * sin ( angle / 3 );

    glutPostRedisplay ();
}

int main ( int argc, char * argv [] )
{
                                // initialize glut
    glutInit            ( &argc, argv );
    glutInitDisplayMode ( GLUT_DOUBLE | GLUT_RGBA | GLUT_DEPTH | GLUT_STENCIL );
    glutInitWindowSize  ( 500, 500 );


                                // create window
    glutCreateWindow ( "OpenGL stencil shadow volume demo (z-fail)" );

                                // register handlers
    glutDisplayFunc  ( display );
    glutReshapeFunc  ( reshape );
    glutKeyboardFunc ( key     );
    glutIdleFunc     ( animate );

    init ();

    initExtensions ();

    ambMap      = createTexture2D ( true,  "../../Textures/oak.bmp" );
    diffuseMap  = createTexture2D ( true,  "../../Textures/Oxidated.jpg" );

    glutMainLoop ();

    return 0;
}
